"""
;
;   MODULE attofile.py
;
;   DESCRIPTION This file is responsible for servicing file read and write
;   requests (file download/uploads).
;
;   THIS PROGRAM AND THE INFORMATION CONTAINED HEREIN IS THE PROPERTY OF
;   ATTO TECHNOLOGY, INC. AND SHALL NOT BE REPRODUCED, COPIED, OR USED IN
;   WHOLE OR IN PART OTHER THAN AS PROVIDED FOR IN THE LICENSE AGREEMENT
;   PURSUANT TO WHICH IT WAS FURNISHED.
;
;   COPYRIGHT (c) ATTO TECHNOLOGY, INC. 2019-2023
;   ALL RIGHTS RESERVED.
;
"""
import sys
import getopt
import os
import tarfile
import re
import struct
import datetime
import json
import shutil
import threading
import time
import binascii
from ctypes import *
from __builtin__ import True, False
from pickle import TRUE
import logging
import subprocess
import zipfile
logger = None

TAR_CMD_LOC = "/usr/bin/tar"

# File and file extension defines
ZBD_EXT = ".zbd"
BOOT_BLOCK_EXT = "*.bin"
UBOOT_EXT = "*.bin"
DTB_EXT = "*.dtb"
KERNEL_EXT = "*.knl.gz"
BOOT_BLOCK_FILE = "atto_bbk.bin"
BOOT_BLOCK_FULL_PATH = ("Uboot/" + BOOT_BLOCK_FILE)
UBOOT_FILE = "uboot_fip.bin"
UBOOT_FULL_PATH = ("Uboot/" + UBOOT_FILE)
DTB_FILE = "fsl-ls1046a-rdb.dtb"
DTB_FILE_DEVKIT = "fsl-ls2088a-rdb.dtb"
KERNEL_FILE = "Image.knl.gz"
APPS_TAR = "apps.tar"
VERSION_FILE = "version.json"
CONFIG_FILE = "config.json"
ROOT_FS_FILE = "rootfs.tar.gz"
DEFAULT_FILE_LOC = "/home/admin/apps/"
SYSTEM_FILE_LOC = DEFAULT_FILE_LOC + "system/"
DEFAULT_MFG_FILE_LOC = "/tmp"
FLASH_MAP_FILE = "flashmap.json"
SAS_FW_FILE = "8100Tfw.rom"
STORCLI_FILE = "storcli.py"
GUI_FILE = "gui.tar.gz"
GUI_ZIP_NAME = "web.zip"
GUI_FILE_LOC = DEFAULT_FILE_LOC + "webgui/"
GUI_ZIP_LOC = DEFAULT_FILE_LOC + "webgui/web/"
APPS_TAR_LOC = "/usr/apps_tar_%c/"
INSTALL_FNAME = "install_required"
INSTALL_ROOTFS_FNAME = "install_rootfs_required"
SUCCESS_FNAME = "/tmp/zbd_success"
FAILURE_FNAME = "/tmp/zbd_failure"

## File titles for ZBD pieces as they appear in the version.json file
VER_FILE_BUILDTYPE = "buildtype"
VER_FILE_PRODID = "productid"
VER_FILE_PHYTYPE = "phytype"
VER_FILE_SASFW = "sasfirmware"
VER_FILE_GUI = "guiversion"

# MTDBlock names (come from DTB/flash map file).
# Names preceded by '%s -' are part of dual boot
# and will be preceded by the boot side indicator (A or B).
MTD_FLASH_CHIP_NAME = "Flash Chip"
MTD_STR_BOOT_BLOCK = "Boot Block"
MTD_STR_UBOOT = "Uboot"
MTD_STR_DTB = "DTB"
MTD_STR_KERNEL = "Kernel"

METADATA_UNKNOWN_ADDR = "0xFFFFFFFFF"
METADATA_UNKNOWN_SIZE = "0x000000000"

BOOT_SIDE_A = "A"

# Template for opening mtd block files
MTD_BLOCK_LOC_TEMPLATE = "/dev/mtdblock%d"

SCRIPT_USAGE = "Usage: python attofile.py [-p filepath | -g file]"
VALID_PUT_EXTS = [ZBD_EXT]
VALID_GET_EXTS = []
SCRIPT_DESC = """
This tool is used as a means to upload and download items to and
from the XC8100T product. Use the -p option to upload a file to the
board.  Use the -g option to download a file from the board.
Valid file extensions:
Put: %s
Get: %s
""" % (
    VALID_PUT_EXTS,
    VALID_GET_EXTS,
)

# Common file locations
FILE_PROC_MTD = "/proc/mtd"

# Print codes are two digits: [Severity Level][FTP Passthrough]
# Example: Code 31 means an error message that should be passed through to FTP

# Print severity level codes
PRINT_SEV_DEBUG = 0
PRINT_SEV_INFO = 1
PRINT_SEV_WARNING = 2
PRINT_SEV_ERROR = 3

# FTP Passthrough Flag
PRINT_FTP_PASS = 1
PRINT_FTP_NO_PASS = 0

# How long we wait before printing progress dots
PROGRESS_DOT_TIME = 8

# Set up a working directory
ATTOFILE_DIR_NAME = "attofile"
working_dir_path = None

# This dictionary describes the list of MTD blocks
# {MTDBlock.name : MTDBlock Object}
# It is populated after parsing /proc/mtd
mtd_blocks = {}

PROGRESS_DOTS = "..."

# overall zbd update progress. Start at non-zero 
progress = 5

# If this is set to true, all images will be updated regardless
# of their versions.  Used when an incoming ZBD's flash map is
# different, or can be set to to true manually for debugging/
# development purposes.
# Manufacturing builds always update all components.
# We don't have versions nailed down for 8100T yet, so until we do, just update
# everything.
update_all = True

# Set a global write_zbd_side, because we only need to look it up once
# and all components adhere to it. Saving curr_boot_side is because we
# are usually going to be writing to the actual opposite, but if the
# flash map on Chip 1 is increasing, then curr_boot_side is actually the
# same as write_side.

# CMK 03MAR23 -- For the short term, there are no "sides" and there are 
# no "versions". It will take longer to process a ZBD, but this is only
# temporary, until the flash map is sorted out, and we determine what the
# concept of "version" means. 
write_zbd_side = BOOT_SIDE_A #None
curr_boot_side = BOOT_SIDE_A #None

zbd_build_type = None
EXIT_CODE_MISSING_BUILD_TYPE = 31
EXIT_CODE_INVALID_BUILD_TYPE = 32
BUILD_TYPES = ["alpha","beta","engdev","final","mestest","systest","xmode"]

zbd_product_id = None
EXIT_CODE_MISSING_PRODUCT_ID = 33
EXIT_CODE_INVALID_PRODUCT_ID = 34
XC8100T_ATTO_BOARD = "4d"
XC8100T_DEVKIT = "4e"
EXIT_CODE_MISSING_VERSION_JSON = 43
EXIT_CODE_MISMATCH_PRODUCT_ID = 44
MISMATCH_PRODUCT_ID_MSG = ("Mismatch between ZBD product ID (%s) and "
                          "HW product ID (%s)")

EXIT_CODE_ZBD_VALIDATION_FAIL = 38

zbd_phy_type = None
EXIT_CODE_MISSING_PHY_TYPE = 40
EXIT_CODE_INVALID_PHY_TYPE = 39

# The *incoming* ZBD's version.json file's data as a JSON object
version_json = None
EXIT_CODE_VERJSON_EXTRACT_FAIL = 41

zbd_sas_firmware_ver = None
EXIT_CODE_MISSING_SAS_FW = 45
# Failed to extract and run sas fw update
EXIT_CODE_SAS_FW_FAIL = 42

zbd_gui_ver = None
EXIT_CODE_MISSING_GUI_VER = 46
EXIT_CODE_GUI_VER_FAIL = 47

# A list representation of the files that are required to be in a ZBD
# Don't require CONFIG_FILE for backflash.
# TODO: Put CONFIG_FILE back at some point in 2021.
# Manufacturing builds do not have a ZBD component requirement.
required_zbd_pieces = [
    VERSION_FILE,  # Versions file, must have this name
    FLASH_MAP_FILE,  # Flash map description file, must have this name
]

# Don't recursively clean up
update_successful = False
starting_clean_up = False

# Set up a completely different workflow if this is a manufacturing
# context where the primary filesystem is being wiped and rewritten. 
manufacturing = False


class FtpError(Exception):
    """An error occurred during FTP

    Attributes:
        message -- explanation of the error
    """

    def __init__(self, message=""):
        self.message = message

class CopyThread(threading.Thread):
    """Copy the data from the input file into the destination file.

    Args:
        dest_path (str): The path to the destination file.
        offset (int): An offset into the file to write to in the
            destination file.
        input_file_path (str): The path to the input file whose contents
            are to be copied.

    Note: If err_message is not None upon thread completion, an IOError
        exception was caught.
    """

    def __init__(self, dest_path, offset, input_file_path):
        threading.Thread.__init__(self)
        self.dest_path = dest_path
        self.offset = offset
        self.input_file_path = input_file_path
        self.err_message = None

    def run(self):
        # Must use 'rb+' here for the flash file or else
        # the data before the 'seek' will get zeroed out
        SIZE_4K = 0x1000
        try:

            with open(self.dest_path, "rb+") as dest_file:

                dest_file.seek(self.offset)
                with open(self.input_file_path, "rb") as input_file:

                    # Continually write data to flash using 4K chunks
                    data = input_file.read(SIZE_4K)
                    while data:
                        dest_file.write(data)
                        data = input_file.read(SIZE_4K)

        except OSError as err:
            self.err_message = (
                "Error attempting to write %s, err: %s"
                % (self.input_file_path, err),
                PRINT_SEV_DEBUG,
            )
        except IOError as ioe:
            self.err_message = (
                "Error attempting to write %s, err: %s"
                % (self.input_file_path, ioe),
                PRINT_SEV_DEBUG,
            )


class MTDBlock:
    """This class represents an entry in the MTD block table which is located
    at /proc/mtd.

    Attributes:
        mtd_num (int): The block number for this MTD block.
        size (int): The size of this MTD block.
        erase_size (int): The erase_size for this MTD block.
        name (str): The name/description of this MTD block.

    """

    def __init__(self, mtd_num, size, erase_size, name):
        """Initialization function for the MTDBlock class.

        Args:
            mtd_num (int): The block number for this MTD block.
            size (int): The size of this MTD block.
            erase_size (int): The erase_size for this MTD block.
            name (str): The name/description of this MTD block.


        """
        self.mtd_num = mtd_num
        self.size = size
        self.erase_size = erase_size
        self.name = name

    def __repr__(self):
        return 'mtd%d: %08x %08x "%s"' % (
            self.mtd_num,
            self.size,
            self.erase_size,
            self.name,
        )

    def __str__(self):
        return 'mtd%d: %08x %08x "%s"' % (
            self.mtd_num,
            self.size,
            self.erase_size,
            self.name,
        )

def attofile_print(message, sev=PRINT_SEV_INFO, ftp_pass=PRINT_FTP_NO_PASS):
    """Custom print function which standardizes output messages.
    Format: "[sev|ftp_pass] message" where sev and ftp_pass are one-digit
    numerical values.

    Args:
        message (str): The message to be printed to standard output.
        sev (int): Message severity level
        ftp_pass (int): Flag to specify whether or not this message should
            be collected and sent to the FTP client by the FTP server.
    """
    # Do not allow a debug message to go through to the FTP client
    if sev == PRINT_SEV_DEBUG and ftp_pass == PRINT_FTP_PASS:
        print(
            "[%d%d] FTP passthrough flag revoked from next message"
            % (PRINT_SEV_WARNING, PRINT_FTP_NO_PASS)
        )
        ftp_pass = PRINT_FTP_NO_PASS

    print("[%d%d] %s" % (sev, ftp_pass, message))

    # By default, the prints to stdout are buffered when not connected
    # to a tty.  This means that from the OS' perspective,
    # the program has not written anything to stdout until either the
    # buffer is full or you manually flushed the output.
    # Thus, make sure to flush here so that the calling program (bftpd)
    # can get updates in real-time.
    sys.stdout.flush()

def update_progress(inc):
    """Update the progress file 'fwupdating' with a value
    representing the % complete. Saturate the percent complete
    to 99.

    Args:
        An increment amount
    """
    global progress

    # don't allow progress to grow past 99
    if progress + inc > 99:
        progress = 99
    else:
        progress += inc
    
    try:
        f = open("/home/admin/ftp/fwupdating", "w")
        f.write("%d" % progress)
        f.close()
    except:
        print("unable to update status")
    



def report_error(error_code, message, print_usage=False):
    """Print the given error message (and optionally the usage) and exit
    the script.

    Args:
        message (str): The error message to print to the user.
        print_usage (bool, optional): Specifies whether or not to print usage
            text, defaults to false.

    """
    attofile_print(message, PRINT_SEV_ERROR, PRINT_FTP_PASS)
    if print_usage:
        attofile_print(SCRIPT_USAGE, PRINT_SEV_INFO, PRINT_FTP_NO_PASS)

    clean_up()

    sys.exit(error_code)


def report_error_no_cleanup(error_code, message, print_usage=False):
    """Print the given error message (and optionally the usage) and exit
    the script. Don't do any file dir or cleanup.

    Args:
        message (str): The error message to print to the user.
        print_usage (bool, optional): Specifies whether or not to print usage
            text, defaults to false.

    """
    attofile_print(message, PRINT_SEV_ERROR, PRINT_FTP_PASS)
    if print_usage:
        attofile_print(SCRIPT_USAGE, PRINT_SEV_INFO, PRINT_FTP_NO_PASS)

    sys.exit(error_code)


def parse_proc_mtd():
    """Retrieve the contents of the /proc/mtd linux file and fill out a global
    list of MTD block information using the data parsed from said file.

    Args:
        None

    """

    col_header_pat = re.compile(r"^dev:\s+size\s+erasesize\s+name$")
    mtd_entry_pat = re.compile(
        r"^mtd(\d+): ([0-9a-f]{8}) ([0-9a-f]{8}) \"(.+)\"$"
    )

    with open(FILE_PROC_MTD, "r") as proc_mtd_file:

        # Go line by line and extract MTD Block information
        attofile_print("/proc/mtd:", PRINT_SEV_DEBUG)
        for line in proc_mtd_file.readlines():

            logger.info(line.rstrip())
            attofile_print(line.rstrip(), PRINT_SEV_DEBUG)

            # Skip over the first line which just contains the column headers
            if col_header_pat.match(line):
                continue

            mtd_info_matches = mtd_entry_pat.match(line)
            if not mtd_info_matches:
                logger.error("Unable to process /proc/mtd line: %s" % line)
                report_error(1, "Unable to process /proc/mtd line: %s" % line)

            # Add this mtd block to the global dictionary
            # {MTDBlock Name: MTDBlock object}
            mtd_blocks[mtd_info_matches.group(4)] = MTDBlock(
                int(mtd_info_matches.group(1)),  # mtd_num
                int(mtd_info_matches.group(2), base=16),  # size
                int(mtd_info_matches.group(3), base=16),  # erasesize
                mtd_info_matches.group(4),
            )  # name


def initialize_working_directory(file_path):
    """Create a working directory at the same level as the passed in
    file.  If it already exists, clear it out.

    Args:
        file_path (str): The file path for the file passed into this
            script.  The resulting working directory will exist at the
            same level as the file.

    Returns:
        bool: True if working directory was successfully created, false
            otherwise.
    """
    global working_dir_path

    path = os.path.split(file_path)[0]
    if path:
        working_dir_path = os.path.join(path, ATTOFILE_DIR_NAME)
    else:
        working_dir_path = ATTOFILE_DIR_NAME

    if os.path.isdir(working_dir_path):
        # If the directory already exists, delete everything in it
        attofile_print(
            "Working directory already exists, deleting remnants",
            PRINT_SEV_WARNING,
        )
        clean_up(completed=False)
    else:
        os.mkdir(working_dir_path)

    return True


def validate_zbd_contents(zbd_path):
    """Create a tarfile object for this ZBD file and validate its contents.

    Args:
        zbd_path (str): The path to the ZBD file to be opened and validated.

    Returns:
        bool: True if ZBD was successfully processed and validated, false
            otherwise.
    """
    zbd_good = True

    # First create the tarfile object which encompasses the passed in ZBD file
    try:
        tar_zbd = tarfile.open(name=zbd_path, mode="r")
    except tarfile.TarError:
        attofile_print(
            "TarError encountered when calling open on tarfile",
            PRINT_SEV_DEBUG,
        )
        return False

    # Iterate through each item in the incoming ZBD and ensure it matches up
    # to something in the required zbd pieces list.
    zbd_members = tar_zbd.getmembers()
    file_tracker = []
    for member_info in zbd_members:
        file_name = member_info.name

        # Check the file_name against all of the pieces in the required list
        found_match = False
        for zbd_piece in required_zbd_pieces:

            if file_name == zbd_piece:
                # Exact match
                found_match = True

                if file_name in file_tracker:
                    # Um...we already have this file...
                    attofile_print(
                        "Multiple files named %s found in ZBD" % zbd_piece,
                        PRINT_SEV_ERROR,
                    )
                    zbd_good = False
                    break

                # Make a note that we received this piece
                file_tracker.append(zbd_piece)
                break

            # If this is a file which can have any base name, compare exts
            if zbd_piece.startswith("*"):

                file_ext = os.path.splitext(file_name)[1]
                zbd_piece_ext = os.path.splitext(zbd_piece)[1]
                if file_ext == zbd_piece_ext:
                    found_match = True
                    # Make a note that we received this piece
                    file_tracker.append(zbd_piece)
                    break

        # Did we find a match throughout that process?
        if not found_match:
            attofile_print(
                "Unknown file found in ZBD: %s" % file_name, PRINT_SEV_WARNING
            )
            # Don't add this file to the file_tracker

    # Let's review to see if any pieces are missing
    for zbd_piece in required_zbd_pieces:
        if zbd_piece not in file_tracker:
            # ZBD was missing stuff!!!
            attofile_print("ZBD missing file %s" % zbd_piece, PRINT_SEV_ERROR)
            zbd_good = False

    # TODO We should probably inspect the app tarball more closely and
    # validate it as well, but I'm not sure what that looks like just yet.

    # Ensure that we close what we opened
    tar_zbd.close()

    return zbd_good


def create_raw_data_flash_map_file(json_file_path):
    """
    Extracts the data from the given json file and creates a new
    file which contains the information (name, address, size) in raw
    form.

    Args:
        json_file_path (str): Path to a json flash map file.

    Returns:
        str: The name of the newly created file or 'None' if an
            error occurred.
    """

    def convert_to_raw_data_pieces(dict_list, name):
        """Helper function that searches a list of dictionaries to find
        a dictionary whose 'Name' element matches the given name.  It then
        packs the other pieces of the dictionary, 'Chip', 'Addr' and 'Size',
        into a writable string and returns those pieces.

        Args:
            dict_list (list[dict]): A list of dictionaries which represent
                the new flash map information.
            name (str): The MTD Block name to search for in the dictionary
                list.

        Returns:
            tuple(str, str, str): Strings representing the chip, address and
                size of the flash pieces corresponding to the given name, or
                'None' if an error occurred.  These strings are produced as the
                result of a struct.pack call.
        """
        try:
            image_dict = [dict for dict in dict_list if dict["Name"] == name][
                0
            ]
            if not image_dict:
                attofile_print(
                    "%s not found in incoming flash map" % name,
                    PRINT_SEV_ERROR,
                )
                return None

            try:
                chip = image_dict["Chip"]
            except KeyError:
                # Prior to PR and SEU, there only used to be Chip 1.
                attofile_print("No Chip element found, using default flash chip")
                chip = MTD_FLASH_CHIP_NAME

            addr = struct.pack("<I", int(image_dict["Addr"], 16))
            size = struct.pack("<I", int(image_dict["Size"], 16))
            return (chip, addr, size)

        except KeyError as ke:
            attofile_print(
                "Encountered KeyError with %s, err: %s" % (name, ke),
                PRINT_SEV_ERROR,
            )
            return None

        except IndexError as ie:
            # This can occur if the current flash map has a component that
            # does not exist in the ZBD that is being flashed on (which is
            # usually a back-flash situation). It won't be used once the
            # system is rebooted, so just use something preposterous that
            # won't actually break anything.
            attofile_print(
                "Encountered IndexError with %s, err: %s" % (name, ie),
                PRINT_SEV_ERROR,
            )
            return (
                MTD_FLASH_CHIP_NAME,
                METADATA_UNKNOWN_ADDR,
                METADATA_UNKNOWN_SIZE,
            )

        # Shouldn't get here
        return None

    new_file_name = "raw_flash_map.bin"

    # Read the contents of the json data into a json object
    with open(json_file_path) as json_file:
        try:
            json_data = json.load(json_file)
        except ValueError as val_err:
            # There is a more specific JSONDecodeError but it isn't
            # available in python 2.
            attofile_print(
                "Parsing json data in flashmap.json failed with message:\n"
                "     %s" % val_err.message,
                PRINT_SEV_ERROR,
            )
            return None

    # mtdblocks is the only top level node
    json_data = json_data["mtdblocks"]

    # Create our new file and put the flash map addresses and sizes in.
    dir_path = os.path.split(json_file_path)[0]
    with open(os.path.join(dir_path, new_file_name), "wb") as raw_data_file:

        # **NOTE** The order in which these pieces go in MUST match that of
        # the "flash_map" struct found in the flash_properties.h file of the
        # adsa/kernel repository.

        # The chip name is returned by convert_to_raw_data_pieces because
        # some entities need to know which chip is which. However, the flash
        # map is only consumed by uboot, which doesn't care about the chips.

        # Boot Block
        boot_block_chip, boot_block_addr, boot_block_size = convert_to_raw_data_pieces(
            json_data, MTD_STR_BOOT_BLOCK
        )
        raw_data_file.write(boot_block_addr)
        raw_data_file.write(boot_block_size)

        # U-Boot
        uboot_chip, uboot_addr, uboot_size = convert_to_raw_data_pieces(
            json_data, MTD_STR_UBOOT
        )
        raw_data_file.write(uboot_addr)
        raw_data_file.write(uboot_size)

        # DTB
        dtb_a_chip, dtb_a_addr, dtb_a_size = convert_to_raw_data_pieces(
            json_data, MTD_STR_DTB
        )
        raw_data_file.write(dtb_a_addr)
        raw_data_file.write(dtb_a_size)

        # Kernel
        kernel_a_chip, kernel_a_addr, ker_a_sz = convert_to_raw_data_pieces(
            json_data, MTD_STR_KERNEL
        )
        raw_data_file.write(kernel_a_addr)
        raw_data_file.write(ker_a_sz)

    attofile_print("Wrote flashmap %s" % new_file_name, PRINT_SEV_DEBUG)

    return new_file_name


def commit_to_flash(file_path, flash_chip, flash_addr):
    """Write the given file to the flash by accessing the "Flash Chip
    MTDBlock and writing the given amount to the given address.

    Args:
        file_path (str): The path to the file to write to flash.
        flash_chip (str): The name of the flash, e.g. MTD_FLASH_CHIP_NAME
        flash_addr (int): Address to write the file to in flash.

    Returns:
        True if file was successfully written to flash, False otherwise.
    """

    logger.info("Writing to chip %s at addr %x" % (flash_chip, flash_addr))
    attofile_print(
        "Writing to chip %s at addr %x" % (flash_chip, flash_addr),
        PRINT_SEV_DEBUG,
    )

    # Make sure we can get the "Flash Chip" mtd block
    flash_chip_path = None
    for mtd_block in mtd_blocks:
        if mtd_block == flash_chip:
            flash_chip_path = (
                MTD_BLOCK_LOC_TEMPLATE % mtd_blocks[mtd_block].mtd_num
            )
            break

    if not flash_chip_path:
        logger.error("Unable to find MTD Block for flash chip")
        attofile_print(
            "Unable to find MTD Block for flash chip", PRINT_SEV_ERROR
        )
        return False

    logger.debug("Flash chip path: %s" % flash_chip_path)

    # Start a CopyThread thread which writes the input file to the
    # flash chip at the given flash address
    logger.debug("Start a CopyThread(%s, 0x%x, %s)"
                 % (flash_chip_path, flash_addr, file_path))
    copy_thread = CopyThread(flash_chip_path, flash_addr, file_path)
    copy_thread.start()

    # While the write is occurring, print out 'in progress' dots
    # every so often to let the user know we're still working
    progress_dot_time = time.time()
    while copy_thread.isAlive():
        if (time.time() - progress_dot_time) > PROGRESS_DOT_TIME:
            logger.info(PROGRESS_DOTS)
            attofile_print(PROGRESS_DOTS, PRINT_SEV_INFO, PRINT_FTP_PASS)
            progress_dot_time = time.time()
            update_progress(3)

    # Check to make sure there is no error message
    if copy_thread.err_message:
        logger.error("Copy thread encountered an error %s"
                     % copy_thread.err_message)
        attofile_print(copy_thread.err_message, PRINT_SEV_ERROR)
        return False

    return True


def save_apps_and_version_files(tar_zbd, write_path):
    """Save the apps.tar file.  The apps.tar file is used
    by the installer app at next reboot. Additionally save off the
    version.json, since the install_app.py script uses it to know which
    installers to run from apps.tar.

    Args:
        tar_zbd (tarfile object): The tar_zbd file to extract.
        write_path (string) : The path where apps are to be written.
    """
    direxists = os.path.isdir(write_path)
    if not direxists:
        try:
            # make the directory
            os.mkdir(write_path)
            direxists = True
            logger.info("Directory %s successfully created" % write_path)
        except OSError:
            logger.error("Unable to create directory %s" % write_path)
            report_error(7, "Unable to create apps tar directory")

    if direxists:
        # delete the old apps tar for safety's sake.
        try:
            os.remove(write_path + APPS_TAR)
            logger.info("Old apps.tar file deleted")
        except OSError:
            logger.info("No existing apps.tar file found")
            attofile_print("No existing apps tar file", PRINT_SEV_DEBUG)

        try:
            # got an error using trailing slash in the path.
            tar_zbd.extract(APPS_TAR, path=write_path[:-1])
            logger.info("New apps.tar extracted from ZBD and placed in %s"
                        % write_path[:-1])
        except TarError:
            logger.error("New apps.tar could not be written to %s"
                         % write_path[:-1])
            report_error(8, "New applications could not be written.")

        # delete the old version.json file
        try:
            os.remove(write_path + VERSION_FILE)
            logger.info("Old version.json file deleted")
        except OSError:
            logger.info("No existing version.json file found")
            attofile_print("No existing version.json file", PRINT_SEV_DEBUG)

        try:
            tar_zbd.extract(VERSION_FILE, path=write_path[:-1])
            logger.info("New version.json extracted from ZBD and placed in %s"
            % write_path[:-1])

        except TarError:
            logger.error("New version.json could not be written to %s"
                         % write_path[:-1])
            report_error(8, "New version.json file could not be written.")

    else:
        logger.error("New apps and version.json could not be written. "
                     "Bath path.")
        report_error(9, "New apps and version.json could not be written. "
                     "Bad path.")


def save_off_rootfs(tar_zbd, write_path):
    """Save off the rootfs.tar.gz file to be processed by the installer app at
    next boot.

    Args:
        tar_zbd (tarfile object): The tar_zbd file to extract.
        write_path (string) : The path where rootfs.tar.gz is to be written.
    """
    direxists = os.path.isdir(write_path)
    if not direxists:
        try:
            # make the directory
            os.mkdir(write_path)
            direxists = True
            logger.info("Directory %s successfully created" % write_path)
        except OSError:
            logger.error("Unable to create directory %s" % write_path)
            report_error(7, "Unable to create rootfs tar directory")

    if direxists:
        # delete the old rootfs.tar.gz for safety's sake.
        try:
            os.remove(write_path + ROOT_FS_FILE)
            logger.info("Old rootfs.tar.gz file deleted")
        except OSError:
            logger.info("No existing rootfs.tar.gz file found")
            attofile_print("No existing rootfs.tar.gz file", PRINT_SEV_DEBUG)

        try:
            # got an error using trailing slash in the path.
            tar_zbd.extract(ROOT_FS_FILE, path=write_path[:-1])
            logger.info("New rootfs.tar.gz extracted from ZBD and placed in %s"
                        % write_path[:-1])
        except TarError:
            logger.error("New rootfs.tar.gz could not be written to %s"
                         % write_path[:-1])
            report_error(8, "New rootfs.tar.gz could not be written.")

    else:
        logger.error("New rootfs.tar.gz could not be written. Bath path.")
        report_error(9, "New rootfs.tar.gz could not be written. Bad path.")


# Extract install_app from the tar file
def extract_install_app_from_tar(src_path):

    """Extract the installer_app from the apps.tar file
    ARGS: src_path = /usr/apps_tar_<side>
    """

    src_apps_file = src_path + APPS_TAR
    dst_apps_dir = DEFAULT_FILE_LOC

    # extract install_app.tar, and inst_install_app.sh to dst_apps_dir

    str = (
        "tar -xf "
        + src_apps_file
        + " -C "
        + dst_apps_dir
        + " inst_install_app.sh install_app.tar"
    )

    try:
        attofile_print("Extracting: %s" % str, PRINT_SEV_DEBUG)
        logger.info("Extracting: %s" % str)
        os.system(str)

    except Exception:
        logger.error("Failed to extract %s" % str)
        attofile_print("Failed to extract %s" % str)

    # Change directory to /home/admin/apps to run inst_install_app.sh, then
    # back to our current working directory before returning
    cwd = os.getcwd()
    os.chdir(dst_apps_dir)
    # Extract contents of install_app.tar and make any changes to the
    # attoinstall.service unit file prior to our next reboot when the installer
    # app will run.
    os.system("./inst_install_app.sh")
    os.chdir(cwd)


def create_install_required_flag(write_path, install_flag_file):
    """Create an install_required file to indicate to the installer app that it
    has work to do at next boot. If install_required is created, the installer
    app will run and upack the apps.tar bundle to /home/admin/apps. If
    install_rootfs_required is created, the installer app will process the
    incoming rootfs.tar.gz

    Args:
        write_path (string): The path where apps or rootfs are/is to be written.
        install_flag_file (string): File indicating that the install app has
                                    work to perform at next boot.
    """

    stat = os.path.isdir(write_path)
    if stat:
        try:
            logger.info("Attempting to open %s" % (write_path + install_flag_file))
            with open(write_path + install_flag_file, "w") as fil:
                fil.write("1\n")
        except IOError:
            logger.error("Unable to open %s" % (write_path + install_flag_file))
            stat = False

    if not stat:
        logger.error("Could not signal installer app to process updates on boot."
                     " Directory %s does NOT exist" % write_path)
        report_error(
            10,
            "Could not signal installer app to process updates on boot. \
                     Firmware update incomplete.",
            PRINT_SEV_ERROR,
            PRINT_FTP_PASS,
        )


def crc32_from_file(filename):
    """Generate a 32-bit unsigned CRC value across a given input file

    Args:
        filename (string) : The filename (with path) from which the CRC
        is calculated.
    """

    buf = open(filename, "rb").read()
    buf = binascii.crc32(buf) & 0xFFFFFFFF
    return "%08X" % buf

# Note that Manufacturing builds do not have any versions or 
# boot side checking, side A is unconditionally updated.
def commit_zbd(
    zbd_path, 
    write_side, 
):
    """Extract the contents of the ZBD file piece by piece and service
    them appropriately.

    Args:
        zbd_path (str): The path to the ZBD file.
        write_side (char): The side to write the ZBD contents into in
            flash.
    """
    global working_dir_path
    global update_all
    global curr_boot_side
    global update_successful
    global manufacturing

    update_progress(0)

    # First we need to grab the flash map for the incoming files
    try:
        tar_zbd = tarfile.open(name=zbd_path, mode="r")
        logger.info("Untarred the ZBD file successfully")
    except tarfile.TarError:
        logger.error("TarError encountered when calling open on ZBD tarfile")
        report_error(12, "TarError encountered when calling open on tarfile")

    # Get the flash map file out of the zbd
    filenames = tar_zbd.getnames()
    if FLASH_MAP_FILE not in filenames:
        logger.error("Flash Map file not found in ZBD")
        report_error(13, "Flash Map file not found in ZBD")

    for filename in filenames:
        if filename == FLASH_MAP_FILE:
            new_flash_map_filename = filename
            break

    logger.info("Attempt to convert flash map file to JSON object")
    # Convert the flash map file contents into a json object
    tar_zbd.extract(new_flash_map_filename, path=working_dir_path)
    new_flash_map_full_path = os.path.join(
        working_dir_path, new_flash_map_filename
    )
    with open(new_flash_map_full_path) as new_flash_map_file:
        try:
            new_flash_map = json.load(new_flash_map_file)
            logger.info("Flash map JSON object successfully created")
        except ValueError as val_err:
            # There is a more specific JSONDecodeError but it isn't
            # available in python 2.
            logger.error("Parsing json data in flashmap.json failed with "
                         "message: %s" % val_err.message)
            attofile_print(
                "Parsing json data in flashmap.json failed with message:\n"
                "     %s" % val_err.message,
                PRINT_SEV_ERROR,
            )
            tar_zbd.close()
            logger.error("Invalid ZBD Received")
            report_error(14, "Invalid ZBD Received")

    # All the good stuff is within the mtdblocks section
    new_flash_map = new_flash_map["mtdblocks"]

    # Decide the correct DTB to expect
    if zbd_product_id == XC8100T_ATTO_BOARD:
        dtb_file_name = DTB_FILE
    elif zbd_product_id == XC8100T_DEVKIT:
        dtb_file_name = DTB_FILE_DEVKIT
    else:
        tar_zbd.close()
        logger.error("Invalid Product ID Received for DTB Name Decode")
        report_error(15, "Invalid ZBD Received")

    # Iterate through each item in the incoming ZBD and place it in the
    # correct location in flash (assuming the version has changed)
    zbd_members = tar_zbd.getmembers()
    logger.debug("ZBD members: %s" % zbd_members)
   
    for member_info in zbd_members:

        update_progress(5)

        file_name = member_info.name
        logger.info("Processing File: %s" % file_name)
        attofile_print("Processing File: %s" % file_name, PRINT_SEV_DEBUG)

        # Various items that will tell us what
        # to do at the bottom of this loop
        need_extraction = True
        flash_map_string = None

        # Handle this file based on name or extension (if extension, remove
        # the asterisk from the string before comparing; i.e., [1:])

        # Applications
        if file_name == APPS_TAR:
            if manufacturing:
                logger.info("Manufacturing mode, so extracting apps to %s"
                            % DEFAULT_MFG_FILE_LOC)
                attofile_print("Extracting applications for later expansion",
                               PRINT_SEV_INFO, PRINT_FTP_PASS)
                tar_zbd.extract(file_name, path=DEFAULT_MFG_FILE_LOC)
            else:
                # This is a tarball of applications.
                # the apps_tar gets written to a directory based on write side
                apps_write_path = APPS_TAR_LOC % write_side.lower()

                logger.info("Apps write path: %s" % apps_write_path)
    
                # If there is an existing apps tarball on the board, copy it
                # to a rollback location.

                attofile_print(
                    "Storing applications for update on reboot",
                    PRINT_SEV_INFO,
                    PRINT_FTP_PASS,
                )
        
                logger.debug("Calling save_apps_and_version_files()")
                save_apps_and_version_files(tar_zbd, apps_write_path)
    
                # extract the install app from the apps.tar file to ensure the
                # latest version is used during the apps installation after
                # reboot
                logger.debug("Calling extract_install_app_from_tar(%s)"
                             % apps_write_path)
                extract_install_app_from_tar(apps_write_path)
    
                logger.debug("Calling create_install_required_flag(%s, %s)"
                             % (apps_write_path, INSTALL_FNAME))
                create_install_required_flag(apps_write_path, INSTALL_FNAME)

            # Apps are not written to flash and therefore
            # nothing more needs to be done
            continue

        elif file_name == BOOT_BLOCK_FULL_PATH:
            flash_map_string = MTD_STR_BOOT_BLOCK

        elif file_name == UBOOT_FULL_PATH:
            flash_map_string = MTD_STR_UBOOT

        # DTB (*.dtb)
        elif file_name == dtb_file_name:
            flash_map_string = MTD_STR_DTB

        # Linux Kernel (*.knl)
        elif file_name == KERNEL_FILE:
            flash_map_string = MTD_STR_KERNEL

        # SAS FW Files
        elif file_name == SAS_FW_FILE:
            tar_zbd.extract(file_name, path=DEFAULT_FILE_LOC)
            logger.info("Sas Controller FW extracted from ZBD and placed in %s"
            % DEFAULT_FILE_LOC)
            attofile_print("Sas Controller FW extracted from ZBD and placed in %s"
            % DEFAULT_FILE_LOC, PRINT_SEV_INFO,
            )
            continue

        elif file_name == STORCLI_FILE:
            tar_zbd.extract(file_name, path=SYSTEM_FILE_LOC)
            logger.info("storcli.py extracted from ZBD and placed in %s"
            % SYSTEM_FILE_LOC)
            attofile_print("storcli.py extracted from ZBD and placed in %s"
            % SYSTEM_FILE_LOC, PRINT_SEV_INFO,
            )
            continue
        elif file_name == ROOT_FS_FILE:
            # The install app handles installing both the apps and rootfs
            # updates. The rootfs.tar.gz from an incoming ZBD is written to
            # /usr/apps_tar_%c for later processing along with apps.tar.
            rootfs_write_path = APPS_TAR_LOC % write_side.lower()

            logger.info("Rootfs write path: %s" % rootfs_write_path)

            attofile_print(
                "Storing rootfs for update on reboot",
                PRINT_SEV_INFO,
                PRINT_FTP_PASS,
            )
    
            logger.debug("Calling save_off_rootfs()")
            save_off_rootfs(tar_zbd, rootfs_write_path)

            logger.debug("Calling create_install_required_flag(%s, %s)"
                         % (rootfs_write_path, INSTALL_ROOTFS_FNAME))
            create_install_required_flag(rootfs_write_path,INSTALL_ROOTFS_FNAME)
            continue
        elif file_name == GUI_FILE:
            if not GUI_FILE_LOC:
                logger.info("%s not found, creating....... %s"
                             % GUI_FILE_LOC)
                os.makedirs(GUI_FILE_LOC)
            if not GUI_FILE_LOC:
                logger.error("%s not found" % GUI_FILE_LOC)
            else:
                tar_zbd.extract(file_name, path=GUI_FILE_LOC)
                attofile_print("GUI file extracted from ZBD and placed in %s"
                            % GUI_FILE_LOC, PRINT_SEV_INFO)
                logger.info("GUI file (%s) extracted from ZBD and placed in %s"
                            % (file_name,GUI_FILE_LOC))
                output_dir = GUI_FILE_LOC
                Tar_Transform = " --transform=s/" + zbd_gui_ver + "// "
                TAR_CMD = TAR_CMD_LOC + " -C " + GUI_FILE_LOC
                TAR_GZ_LOC = GUI_FILE_LOC + GUI_FILE
                ZIP_LOC = GUI_ZIP_LOC + GUI_ZIP_NAME
                TAR_CMD = TAR_CMD + Tar_Transform + "-zxf " + TAR_GZ_LOC
                #print (TAR_CMD)
                RC = os.system(TAR_CMD + ">/dev/null 2>&1")
                if (RC != 0):
                    logger.error("GUI extract failed with "
                                 "code: %d" % RC)
                else:
                    with zipfile.ZipFile(ZIP_LOC, 'r') as zip_ref:
                        zip_ref.extractall(GUI_ZIP_LOC)
                    try:
                        os.remove(TAR_GZ_LOC)
                        logger.info("%s file deleted" % TAR_GZ_LOC)
                    except OSError:
                       logger.info("%s file not found" % TAR_GZ_LOC)
                       attofile_print("%s file not found" % TAR_GZ_LOC, PRINT_SEV_DEBUG)
                    try:
                        os.remove(ZIP_LOC)
                        logger.info("%s file deleted" % ZIP_LOC)
                    except OSError:
                        logger.info("%s file not found" % ZIP_LOC)
                        attofile_print("%s file not found" % ZIP_LOC, PRINT_SEV_DEBUG)
            continue
        elif file_name == VERSION_FILE:
            # Do nothing since version.json is handled by the APPS_TAR case
            # above.
            logger.info("version.json processed along with apps.tar -- skipping")
            attofile_print(
                "version.json processed along with apps.tar -- skipping",
                PRINT_SEV_INFO,
                PRINT_FTP_PASS,
            )
            continue

        else:
            # Unrecognized file, save it one level above the working
            # directory (because the working directory gets cleaned
            # out at the end) and leave it alone
            logger.warning("Unrecognized file %s extracted to %s" %
                (file_name, DEFAULT_FILE_LOC))
            attofile_print(
                "Extracting unknown file %s to %s"
                % (file_name, DEFAULT_FILE_LOC),
                PRINT_SEV_WARNING,
            )

            tar_zbd.extract(file_name, DEFAULT_FILE_LOC)

            continue

        logger.debug("Flash map string is %s" % flash_map_string)
        attofile_print("Processed %s" % file_name, PRINT_SEV_DEBUG)

        # Check to see if we actually have to perform the update.
        if True:
            logger.info("Performing update...")
            # Did we previously extract the file from the ZBD or no?
            if need_extraction:
                logger.info("Extracting file %s from ZBD tarball to %s" % 
                    (file_name, working_dir_path))
                tar_zbd.extract(file_name, path=working_dir_path)

            # get the full path
            file_path = os.path.join(working_dir_path, file_name)
            logger.debug("Full path to file: %s" % file_path)

            # Get the flash address and size from the flash map
            logger.info("Attempt to find the flash address and size in the "
                        "flash map")
            flash_info_found = False
            flash_chip = None
            flash_addr = None
            mtd_max_size = None
            for image in new_flash_map:
                if image["Name"] == flash_map_string:
                    logger.debug("Matching entry in flash map found")
                    try:
                        flash_chip = image["Chip"]
                    except KeyError:
                        logger.error("No Chip element, defaulting to chip %s"
                                       % MTD_FLASH_CHIP_NAME)
                        attofile_print("No Chip element, defaulting to chip %s"
                                       % MTD_FLASH_CHIP_NAME)
                        flash_chip = MTD_FLASH_CHIP_NAME

                    if flash_chip == "":
                        flash_chip = MTD_FLASH_CHIP_NAME

                    flash_addr = int(image["Addr"], 16)
                    mtd_max_size = int(image["Size"], 16)

                    logger.info("Flash info - chip: %s, flash_addr: 0x%x, "
                                "mtd_max_size: 0x%x"
                                % (flash_chip, flash_addr, mtd_max_size))
                    flash_info_found = True
                    break

            # It would be weird that we don't find flash map
            # info but we should check for it
            if not flash_info_found:
                logger.error("No matching entry for file %s in flash map!"
                             % file_name)
                attofile_print(
                    "No matching entry for file %s in flash map!" % file_name,
                    PRINT_SEV_ERROR,
                )
                report_error(
                    18,
                    "Error while attempting to commit %s file"
                    % os.path.split(file_name)[1],
                )

            # Quick check to make sure that this file stays within the
            # size that was allotted for its mtd block
            if os.stat(file_path).st_size > mtd_max_size:
                logger.error("File %s is larger than its allotted mtd block "
                    "size!" % file_name)
                attofile_print(
                    "File %s is larger than its allotted mtd block size!"
                    % file_name,
                    PRINT_SEV_ERROR,
                )
                logger.error("File size: 0x%08x, MTD Block Size: 0x%08x"
                             % (os.stat(file_path).st_size, mtd_max_size))
                attofile_print(
                    "File size: 0x%08x, MTD Block Size: 0x%08x"
                    % (os.stat(file_path).st_size, mtd_max_size),
                    PRINT_SEV_ERROR,
                )
                logger.error("Error while attempting to commit %s file"
                             % os.path.split(file_name)[1])
                report_error(
                    19,
                    "Error while attempting to commit %s file"
                    % os.path.split(file_name)[1],
                )

            # Write this file to flash
            logger.info("Writing '%s' to flash..." % flash_map_string)
            attofile_print(
                "Writing '%s' to flash..." % flash_map_string,
                PRINT_SEV_INFO,
                PRINT_FTP_PASS,
            )
            logger.debug("Calling commit_to_flash()")
            if not commit_to_flash(file_path, flash_chip, flash_addr):
                logger.error("Error while attempting to commit %s file"
                             % os.path.split(file_name)[1],)
                report_error(
                    20,
                    "Error while attempting to commit %s file"
                    % os.path.split(file_name)[1],
                )

    if (os.path.isfile(DEFAULT_FILE_LOC+SAS_FW_FILE) and
        os.path.isfile(SYSTEM_FILE_LOC+STORCLI_FILE)):
        if zbd_sas_firmware_ver != read_hw_sas_firmware():
            ret = subprocess.call(["python2", SYSTEM_FILE_LOC+STORCLI_FILE, 
                            "upgrade", DEFAULT_FILE_LOC+SAS_FW_FILE])

            if ret > 0:
                logger.error("SAS Firmware update failed to complete, error code: %d" % ret)
                #TODO: Add this later, it will usually fail report_error(42, "Failed to run sas fw update command")
            else: 
                logger.info("SAS Firmware updated successsful!")
                attofile_print(
                "SAS Firmware updated successsful", PRINT_SEV_INFO
                )
        else:
            logger.info("SAS Firmware matches currently installed version, skipping update.")
            attofile_print(
            "SAS Firmware matches currently installed version, skipping update.", PRINT_SEV_INFO
            )
    else: 
        logger.error("Failed to extract and run sas fw update")
        report_error(42, "Failed to extract and run sas fw update")

    logger.info("Closing the ZBD tarfile")
    tar_zbd.close()

    # Updates successful.  Update metadata for this side.
    logger.info("Firmware update successsful!")
    attofile_print(
        "Firmware update successsful, writing metadata...", PRINT_SEV_INFO
    )

    update_progress(100)

    update_successful = True


def clean_up(completed=True):
    """Clean up the workspace area so no remnants are left behind for future
    firmware update processes.

    Args:
        None
    """
    # Navigate to the working directory and delete everything
    global working_dir_path
    global starting_clean_up
    global update_successful
    global manufacturing

    if manufacturing:
        logger.info("Skipping clean up due to manufacturing")
        attofile_print("Skipping clean up due to manufacturing", PRINT_SEV_DEBUG)
        if completed:
            logger.info("ZBD Installation Completed")
            attofile_print("ZBD Installation Completed", PRINT_SEV_DEBUG)
        return
    
    logger.info("Cleaning up working directory")
    attofile_print("Cleaning up working directory", PRINT_SEV_DEBUG)

    if starting_clean_up:
        logger.info("Already tried to clean up")
        attofile_print("Already tried to clean up", PRINT_SEV_DEBUG)
        return
    starting_clean_up = True

    try:
        for root, dirs, files in os.walk(working_dir_path, topdown=False):
            for f in files:
                logger.info("Removing file %s" % f)
                attofile_print("Removing file %s" % f, PRINT_SEV_INFO)
                os.remove(os.path.join(root, f))
            for d in dirs:
                logger.info("Removing dir %s" % d)
                attofile_print("Removing dir %s" % d, PRINT_SEV_INFO)
                os.rmdir(os.path.join(root, d))

        # Remove the working directory itself
        logger.info("Removing working dir at %s" % working_dir_path)
        attofile_print(
            "Removing working dir at %s" % working_dir_path, PRINT_SEV_INFO
        )
        os.rmdir(working_dir_path)

    except OSError as ose:
        logger.error("Error attempting to clean working directory")
        report_error(24, "Error attempting to clean working directory")

    if completed:
        logger.info("Completed")
        if update_successful:
            logger.info("Update Successful")
            if os.path.exists(FAILURE_FNAME):
                logger.info("Removing existing %s file" % FAILURE_FNAME)
                os.remove(FAILURE_FNAME)
            open(SUCCESS_FNAME, "a").close()
            logger.info("Created %s file" % SUCCESS_FNAME)
            print(
                """*************************************
Successfully completed at
%s
*************************************"""
                % datetime.datetime.now()
            )
        else:
            logger.info("Update NOT successful")
            if os.path.exists(SUCCESS_FNAME):
                logger.info("Removing existing %s file" % SUCCESS_FNAME)
                os.remove(SUCCESS_FNAME)
            open(FAILURE_FNAME, "a").close()
            logger.info("Created %s file" % FAILURE_FNAME)
            print(
                """*************************************
Error occurred during processing at
%s
*************************************"""
                % datetime.datetime.now()
            )
    else:
        # Since we haven't completed, and we're just cleaning up 
        # from a previous update, clear the flag indicating that 
        # we're starting the final clean-up.
        starting_clean_up = False


def extract_version_file(zbd_path):
    """This function extracts the version file from the ZBD, converts it to
    a JSON object, and saves it off for later use.

    Args:
        zbd_path (str): The path to the ZBD file
    """
    global version_json

    # Open the tar file
    try:
        tar_zbd = tarfile.open(name=zbd_path, mode="r")
    except tarfile.TarError:
        logger.error("TarError encountered while attempting to open ZBD")
        report_error(
            25,
            "TarError encountered while attempting to open ZBD",
            PRINT_SEV_DEBUG,
        )

    # Pull the versions file out of the ZBD
    try:
       tarinfo_versions = tar_zbd.getmember(VERSION_FILE)
       version_file = tar_zbd.extract(tarinfo_versions, path=working_dir_path)
       tar_zbd.close()
    except KeyError:
        logger.error("Unable to locate version file in ZBD")
        report_error(26, "Unable to locate version file in ZBD")
    except tarfile.TarError:
        logger.error("Failed to extract versions file from ZBD")
        report_error(27, "Failed to extract versions file from ZBD")
    #Convert json data to a python list of dictionaries
    versions_full_path = os.path.join(working_dir_path, VERSION_FILE)
    with open(versions_full_path) as versions_file:
        try:
            version_json = json.load(versions_file)
        except ValueError as val_err:
            # There is a more specific JSONDecodeError but it isn't
            # available in python 2.
            logger.error("Parsing json data in version.json failed with "
                         "message:\n"
                         "     %s" % val_err.message)
            report_error(
               28,
               "Parsing json data in version.json failed with "
               "message:\n"
               "     %s" % val_err.message,
               PRINT_SEV_ERROR,
            )

def read_hw_productid():
    """Reads the productid field from the *existing* version.json file in the
    hardware's rootfs at /home/admin/apps. The productid corresponds to the
    hardware we're running on (i.e., "4d" for a board or "4e" for a devkit).
    Returns the productid.
    """
    version_file_path = os.path.join(DEFAULT_FILE_LOC, VERSION_FILE)
    
    if os.path.isfile(version_file_path):
        with open(version_file_path) as versions_file:
            try:
                version_json = json.load(versions_file)
            except ValueError as val_err:
                logger.error("Parsing json data in %s failed with "
                             "message:\n"
                             "     %s" % (version_file_path, val_err.message))
                report_error(
                   28,
                   "Parsing json data in %s failed with "
                   "message:\n"
                   "     %s" % (version_file_path, val_err.message),
                )

            try:
                product_id = version_json[VER_FILE_PRODID]
            except KeyError:
                product_id = None
                logger.error("Product ID not found in %s" % version_file_path)
                report_error(EXIT_CODE_MISSING_PRODUCT_ID,
                            "Product ID not found in %s" % version_file_path)

            if product_id not in [XC8100T_ATTO_BOARD, XC8100T_DEVKIT]:
                logger.error("Invalid product ID found in %s: %s"
                             % (version_file_path, product_id))
                report_error(EXIT_CODE_INVALID_PRODUCT_ID,
                             "Invalid product ID found in %s: %s"
                             % (version_file_path, product_id))

    else:
        logger.error("Cannot determine HW product ID. File '%s' doesn't exist!"
            % version_file_path)
        report_error(
            EXIT_CODE_MISSING_VERSION_JSON,
            "Cannot determine HW product ID. File %s doesn't exist!"
                % version_file_path
        )

    return product_id

def read_zbd_product_id(version_json):
    """Saves off the product ID of the incoming ZBD to be referenced later and
    validates that the product ID is a type we expect for an XC8100T.

    Args:
        version_json (JSON object): The version.json data as a JSON object.
    """

    global zbd_product_id

    # Get the product type
    try:
        zbd_product_id = version_json[VER_FILE_PRODID]
    except KeyError:
        zbd_product_id = None
        logger.error("Product ID not found in ZBD")
        report_error(EXIT_CODE_MISSING_PRODUCT_ID,
                    "Product ID not found in ZBD")

    if zbd_product_id not in [XC8100T_ATTO_BOARD, XC8100T_DEVKIT]:
        # This ZBD if for a different prouduct! Abort the FW update.
        logger.error("Invalid product ID found in ZBD: %s" % zbd_product_id)
        report_error(EXIT_CODE_INVALID_PRODUCT_ID,
                     "Invalid product ID found in ZBD: %s"
                     % zbd_product_id)

def read_zbd_phy_type(version_json):
    """Saves off the phy type of the incoming ZBD to be referenced later and
    validates that the phy type is a value we expect for an XC8100T.

    Args:
        version_json (JSON object): The version.json data as a JSON object.
    """

    global zbd_phy_type

    # Get the product type
    try:
       zbd_phy_type = version_json[VER_FILE_PHYTYPE]
    except KeyError:
        zbd_phy_type = None
        logger.error("PHY type not found in ZBD")
        report_error(EXIT_CODE_MISSING_PHY_TYPE, "PHY type not found in ZBD")

    # Internal PHYs, external PHYs, or a variant that supports both
    if zbd_phy_type not in ["int", "ext", "both"]:
        # Incorrect phy type for an XC8100T detected!. Abort the FW update.
        logger.error("Invalid phy type found in ZBD: %s" % zbd_phy_type)
        report_error(EXIT_CODE_INVALID_PHY_TYPE,
                     "Invalid phy type found in ZBD: %s" % zbd_phy_type)


def read_zbd_build_type(version_json):
    """Saves the build type of the incoming ZBD to be referenced later and
    validates that the build type is one we expect.

    Args:
        version_json (JSON object): The version.json data as a JSON object.
    """
    global zbd_build_type

    # Get the build type
    try:
        zbd_build_type = version_json[VER_FILE_BUILDTYPE]
    except KeyError:
        zbd_build_type = None
        logger.error("Build type not found in ZBD")
        report_error(EXIT_CODE_MISSING_BUILD_TYPE,
                     "Build type not found in ZBD")

    if zbd_build_type not in BUILD_TYPES:
            # Invalid build type
            logger.error("Invalid build type found in ZBD: %s" % zbd_build_type)
            report_error(EXIT_CODE_INVALID_BUILD_TYPE,
                         "Invalid build type found in ZBD: %s" % zbd_build_type)


def read_hw_sas_firmware():
    """Reads the sasfirmware field from the *existing* version.json file in the
    hardware's rootfs at /home/admin/apps. This is compared to the incoming zbd
    sas firmware version to check if an upgrade/downgrade is needed.
    Returns the sas firmware version.
    """
    version_file_path = os.path.join(DEFAULT_FILE_LOC, VERSION_FILE)
    
    if os.path.isfile(version_file_path):
        with open(version_file_path) as versions_file:
            try:
                version_json = json.load(versions_file)
            except ValueError as val_err:
                logger.error("Parsing json data in %s failed with "
                             "message:\n"
                             "     %s" % (version_file_path, val_err.message))
                report_error(
                   28,
                   "Parsing json data in %s failed with "
                   "message:\n"
                   "     %s" % (version_file_path, val_err.message),
                )

            try:
                sas_firmware_ver = version_json[VER_FILE_SASFW]
            except KeyError:
                sas_firmware_ver = None
                logger.error("SAS Firwmare Version not found in %s" % version_file_path)
                report_error(EXIT_CODE_MISSING_SAS_FW,
                            "SAS Firwmare Version not found in %s" % version_file_path)

    else:
        logger.error("Cannot determine HW product ID. File '%s' doesn't exist!"
            % version_file_path)
        report_error(
            EXIT_CODE_MISSING_VERSION_JSON,
            "Cannot determine HW product ID. File %s doesn't exist!"
                % version_file_path
        )

    return sas_firmware_ver

def read_zbd_sas_firmware_version(version_json):
    """Saves off the SAS controller firmware version of the incoming ZBD to be 
    referenced later to see up a firmware upgrade is needed.

    Args:
        version_json (JSON object): The version.json data as a JSON object.
    """

    global zbd_sas_firmware_ver

    # Get the product type
    try:
        zbd_sas_firmware_ver = version_json[VER_FILE_SASFW]
    except KeyError:
        zbd_sas_firmware_ver = None
        logger.error("SAS Firmware Version not found in ZBD")
        report_error(EXIT_CODE_MISSING_PRODUCT_ID,
                    "SAS Firmware Version not found in ZBD")

def read_zbd_gui_version(version_json):
    """Saves off the GUI version of the incoming ZBD to be
    referenced later when untarring the gui file.

    Args:
        version_json (JSON object): The version.json data as a JSON object.
    """

    global zbd_gui_ver

    # Get the product type
    try:
        zbd_gui_ver = version_json[VER_FILE_GUI]
    except KeyError:
        zbd_gui_ver = None
        logger.error("GUI Version not found in ZBD")
        report_error(EXIT_CODE_MISSING_PRODUCT_ID,
                    "GUI Version not found in ZBD")


def attofile_put(file_path):
    """Responsible for uploading a file and committing it to the
    appropriate location on the XC8100T product.

    Args:
        file_path (str): The path to the file which is to be uploaded
            and committed.
    """
    global curr_boot_side
    global write_zbd_side
    global update_all
    global manufacturing
    global version_json

    print(
        """*************************************
Started at %s
*************************************"""
        % datetime.datetime.now()
    )

    # Verify that the argument leads to a file
    if not os.path.isfile(file_path):
        logger.error("File path is not valid: %s" % file_path)
        report_error(29, "File path is not valid: %s" % file_path)

    logger.info("Accepted file: %s" % os.path.basename(file_path))
    attofile_print(
        "Accepted file: %s" % os.path.basename(file_path),
        PRINT_SEV_INFO,
        PRINT_FTP_PASS,
    )

    if not (initialize_working_directory(file_path)):
        logger.error("Working directory setup failed")
        report_error(30, "Working directory setup failed")
    else:
        logger.info("Working directory: %s" % working_dir_path)
        attofile_print(
            "Working directory: %s" % working_dir_path, PRINT_SEV_DEBUG
        )

    # Populate the MTD block dictionary
    logger.info("Parsing /proc/mtd...")
    parse_proc_mtd()

    # Go do work based on what type of file was sent in
    input_ext = os.path.splitext(file_path)[1].lower()
    if input_ext == ZBD_EXT:
        logger.info("File to put is a ZBD")

        # Extract the version.json file from the ZBD and store it for later use
        extract_version_file(file_path)

        if not version_json:
            # If extracting the version file failed above, we should have
            # errored out and exited, but just in case, exit if we see an
            # unpopulated version_json object. The version_json object is needed
            # to validate the ZBD, so we can't proceed with a firmware update
            # without it.
            logger.error("Unable to extract version.json from ZBD")
            report_error(EXIT_CODE_VERJSON_EXTRACT_FAIL,
                         "Unable to extract version.json from ZBD")

        # Read the product and phy types from the version file of the incoming
        # ZBD. If it's not what we expect, abort the firmware update.
        read_zbd_product_id(version_json)

        if zbd_product_id:
            logger.info("ZBD product ID is: %s" % zbd_product_id)
            attofile_print(
                "ZBD product ID: [%s]" % zbd_product_id,
                PRINT_SEV_INFO,
                PRINT_FTP_PASS,
            )

        # Ensure that the incoming ZBD product ID matches the product ID of the
        # board/devkit onto which it is being flashed. Only proceed with the
        # firmware update on a match.
        hw_product_id = read_hw_productid()

        logger.info("Hardware product ID is: %s", hw_product_id)
        attofile_print(
            "Hardware product ID: [%s]" % hw_product_id,
            PRINT_SEV_INFO,
            PRINT_FTP_PASS,
        )

        if( zbd_product_id != hw_product_id ):

            logger.error(MISMATCH_PRODUCT_ID_MSG
                % (zbd_product_id, hw_product_id))
            report_error(
                EXIT_CODE_MISMATCH_PRODUCT_ID,
                MISMATCH_PRODUCT_ID_MSG % (zbd_product_id, hw_product_id)
            )

        logger.info("Incoming ZBD product ID (%s) matches HW product ID (%s)!"
                    % (zbd_product_id, hw_product_id))
        attofile_print(
            "Incoming ZBD product ID (%s) matches HW product ID (%s)!"
             % (zbd_product_id, hw_product_id),
            PRINT_SEV_INFO,
            PRINT_FTP_PASS,
        )

        read_zbd_sas_firmware_version(version_json)

        if zbd_sas_firmware_ver:
            logger.info("ZBD SAS Firmware Version type is: %s" % zbd_sas_firmware_ver)
            attofile_print(
                "ZBD SAS Firmware Version: [%s]" % zbd_sas_firmware_ver,
                PRINT_SEV_INFO,
                PRINT_FTP_PASS,
            )

        read_zbd_phy_type(version_json)

        if zbd_phy_type:
            logger.info("ZBD PHY type is: %s" % zbd_phy_type)
            attofile_print(
                "ZBD PHY type: [%s]" % zbd_phy_type,
                PRINT_SEV_INFO,
                PRINT_FTP_PASS,
            )

        read_zbd_gui_version(version_json)

        if zbd_gui_ver:
            logger.info("ZBD GUI Version is: %s" % zbd_gui_ver)
            attofile_print(
                "ZBD GUI Version: [%s]" % zbd_gui_ver,
                PRINT_SEV_INFO,
                PRINT_FTP_PASS,
            )

        # Read the incoming build type from the ZBD's version file. If invalid
        # or not found abort the firmware update.
        read_zbd_build_type(version_json)

        if zbd_build_type:
            logger.info("ZBD build type is: %s" % zbd_build_type)
            attofile_print(
                "ZBD build type: [%s]" % zbd_build_type,
                PRINT_SEV_INFO,
                PRINT_FTP_PASS,
            )

        # Validate the given zbd file if the build type is Alpha, Beta, or Final
        if zbd_build_type in ["alpha", "beta", "final"]:
            
            if manufacturing:
                logger.info("ZBD validation skipped for manufacturing")
                attofile_print(
                    "ZBD validation skipped for manufacturing",
                    PRINT_SEV_INFO,
                    PRINT_FTP_PASS,
                )
            else:
                logger.info("Validating ZBD")
                attofile_print("Validating ZBD", PRINT_SEV_INFO, PRINT_FTP_PASS)
                if not validate_zbd_contents(file_path):
                    logger.error("ZBD validation failed")
                    report_error(EXIT_CODE_ZBD_VALIDATION_FAIL,
                                 "ZBD validation failed")
                else:
                    logger.info("ZBD validated successfully")
                    attofile_print(
                        "ZBD validated successfully",
                        PRINT_SEV_INFO,
                        PRINT_FTP_PASS,
                    )

        curr_boot_side = BOOT_SIDE_A
        logger.info("Current boot side: %s" % curr_boot_side)

        # Process ZBD contents
        logger.info("Process ZBD contents (commit_zbd())")
        commit_zbd(
            file_path,
            write_zbd_side,
        )

        # Done with ZBD, delete it
        try:
            os.remove(file_path)
            logger.info("ZBD file deleted successfully at %s" % file_path)
        except OSError:
            logger.error("Unable to delete ZBD file")
            attofile_print("Could not delete ZBD file", PRINT_SEV_WARNING)

    else:
        # Unrecognized file type, return with error
        logger.error("Unrecognized file type: %s" % input_ext)
        report_error(35, "Unrecognized file type: %s" % input_ext)

    # Clean up our workspace
    logger.info("Clean up")
    clean_up()


def main(argv):
    """Verify command line arguments and perform core tasks.

    Args:
        argv (list[str]): List of command line arguments.
    """
    global manufacturing
    global update_all
    global logger

    # Create and configure log file
    logging.basicConfig(filename="/var/log/adsa/attofile.log",
                        format='%(asctime)s %(message)s',
                        filemode='a')
    logger = logging.getLogger()

    # Note: Set level to logging.DEBUG to see all logger messages for debugging
    # purposes.
    logger.setLevel(logging.INFO)
    
    logger.debug("Running attofile.py...")
    
    # Retrieve the command line arguments
    try:
        opts, _ = getopt.getopt(argv, "hp:g:m", ["help"])
    except getopt.GetoptError:
        logger.error("Invalid command line option(s)")
        report_error_no_cleanup(
            0, "Invalid command line option(s)", print_usage=True
        )
        
    # Print usage if help was requested or save the path
    for opt, arg in opts:
        if opt in ("-h", "--help"):
            print(SCRIPT_DESC)
            print(SCRIPT_USAGE)
            print(
                "\nPrint Codes: [SF]\n"
                "    S (Severity): DEBUG=%d, INFO=%d, "
                % (PRINT_SEV_DEBUG, PRINT_SEV_INFO)
                + "WARNING=%d, ERROR=%d\n"
                % (PRINT_SEV_WARNING, PRINT_SEV_ERROR)
                + "    F (FTP Passthrough Flag): NO_PASSTHROUGH=%d, "
                % PRINT_FTP_NO_PASS
                + "PASSTHROUGH=%d\n" % PRINT_FTP_PASS
            )
            logger.debug("Help text requested -- Exiting.")
            sys.exit(0)
        elif opt in ["-p", "-g"]:
            arg_file = arg
            logger.debug("Arg %s used" % arg_file)
        if opt in ("-m"):
            # Manufacturing, force boot side to A, do not validate  
            # anything, and ignore versions.
            print("Enabling manufacturing mode")
            manufacturing = True
            update_all = True
            logger.debug("Manufacturing mode enabled")

    # Can only do one operation per invocation of this command
    arg_list = [arg[0] for arg in opts]
    if "-g" in arg_list and "-p" in arg_list:
        logger.error("Cannot put (-p) and get (-g) a file simultaneously")
        report_error_no_cleanup(
            0, "Only one of options -p and -g allowed", print_usage=True
        )

    # Put this out there in case someone doesn't realize they enabled it
    if update_all:
        logger.info("Note that UPDATE_ALL is enabled")
        attofile_print("UPDATE_ALL IS ENABLED", PRINT_SEV_WARNING)

    # Go to respective operational functions
    input_ext = os.path.splitext(arg_file)[1].lower()
    if "-p" in arg_list:

        # Make sure this is a valid file type for this operation
        if input_ext not in VALID_PUT_EXTS:
            # Remove the file and report the error
            logger.error("The input file's type (%s) is NOT allowed" % input_ext)
            os.remove(arg_file)
            report_error_no_cleanup(
                36, "Invalid input file type for put operation"
            )

        logger.info("Putting file %s" % arg_file)
        attofile_put(arg_file)

    elif "-g" in arg_list:

        # Make sure this is a valid file type for this operation
        if input_ext not in VALID_GET_EXTS:
            logger.error("The input file's type (%s) is NOT allowed" % input_ext)
            report_error_no_cleanup(
                37, "Invalid input file type for get operation"
            )

        logger.info("Getting file %s" % arg_file)
        # TODO attofile_get(arg_file)

    else:
        logger.error("Invalid argument specified: %s" % arg_list)
        report_error_no_cleanup(
            0,
            "Invalid argument specified, how did you get here? "
            "arg_list: %s" % arg_list,
            print_usage=True,
        )

    logger.debug("-- attofile.py DONE --")


if __name__ == "__main__":
    main(sys.argv[1:])
