"""****************************************************************************
;
;   MODULE       storcli.py
;
;   DESCRIPTION  Script to upgrade SAS controller firmware using storcli
;                as well as retrieve controller level details while filtering
;                out unneeded lines
;
;   THIS PROGRAM AND THE INFORMATION CONTAINED HEREIN IS THE PROPERTY OF
;   ATTO TECHNOLOGY, INC. AND SHALL NOT BE REPRODUCED, COPIED, OR USED IN
;   WHOLE OR IN PART OTHER THAN AS PROVIDED FOR IN THE LICENSE AGREEMENT
;   PURSUANT TO WHICH IT WAS FURNISHED.
;
;    USAGE: storcli.py show [firmware | product | serial | phy ]
;           storcli.py show stats <verbose>
;           storcli.py upgrade [firmware_file.rom]"
;
;    RETURNS: 0 on success, nonzero on failure
;
;    NOTES: TODO: Add more functionality as needed, phy commands maybe even
;           plug into new app built to cover missing commands we would have
;           used scrutiny for
;
;   COPYRIGHT (c) ATTO TECHNOLOGY, INC. 2023
;   ALL RIGHTS RESERVED.
;
****************************************************************************"""

import os.path
import sys
import tempfile
import shutil
import subprocess
import shlex

ARG_FUNC_1ST = 1
ARG_FUNC_2ND = 2
ARG_FUNC_3RD = 3

EXEFILE = "/opt/MegaRAID/storcli/storcli64"
CONTROLLER="/c0"
# TODO: Add phy commands and allow for picking phys /px
PHY="/pall"
GET_INFO="show"
STATS="phyerrorcounters"

FW_VERSION="firmware"
PRODUCT_NAME="product"
SERIAL_NUMBER="serial"
PHYS="phys"
PHY_ERR_COUNTERS="stats"
VERBOSE="verbose"

UPGRADE_FIRMWARE ="upgrade"
DL_COMMAND="download file="

FAILURE = False
ERROR_MSG = ""

EXIT_CODE_NO_STORCLI = 1
EXIT_CODE_BAD_ARG = 2
EXIT_CODE_CMD_FAILED = 3

USAGE_TEXT = ("usage : storcli.py show [firmware | product | serial | phys ]\n"
              "        storcli.py show stats <verbose>\n"
              "        storcli.py upgrade [firmware_file.rom]")

# check if the storcli executable exists
try:
    if not os.path.isfile(EXEFILE):
        print "\nERROR: %s : %s : does not exist" % (
            os.path.basename(__file__),
            EXEFILE,
        )
        sys.exit(EXIT_CODE_NO_STORCLI)
    if not sys.argv[ARG_FUNC_1ST]:
        print USAGE_TEXT
        sys.exit(EXIT_CODE_BAD_ARG)
    if sys.argv[ARG_FUNC_1ST] == UPGRADE_FIRMWARE:
        if not os.path.isfile(sys.argv[ARG_FUNC_2ND]):
            print "\nERROR: %s : %s : firmware file does not exist" % (
            os.path.basename(__file__),
            sys.argv[ARG_FUNC_2ND],
            )
            sys.exit(EXIT_CODE_BAD_ARG)
        elif not sys.argv[ARG_FUNC_2ND].endswith('.rom'):
            print "\nERROR: Expected %s to be a .rom file" % (
            sys.argv[ARG_FUNC_2ND],
            )
            sys.exit(EXIT_CODE_BAD_ARG)
except Exception:
    print USAGE_TEXT
    sys.exit(EXIT_CODE_BAD_ARG)

# There are certain assumptions made here such as there is only
# 1 SAS Controller installed.

if sys.argv[ARG_FUNC_1ST] == UPGRADE_FIRMWARE:
    if len(sys.argv) <= 2:
        print USAGE_TEXT
        sys.exit(EXIT_CODE_BAD_ARG)

    fw_file = sys.argv[ARG_FUNC_2ND]
    # /opt/MegaRAID/storcli/storcli64 /c0 download file=<path/to/file>
    cmdline = shlex.split(EXEFILE + " " + CONTROLLER + " " + DL_COMMAND +
                          fw_file + " noverchk")
    try:
        output = subprocess.check_output(cmdline)

        print "\nFirmware flashed successfully"

    except subprocess.CalledProcessError as e:
        for textlin in e.output.split("\n"):
            # ex: Status = Failure
            #     Description = Controller 0 not found
            if FAILURE:
                ERROR_MSG = textlin.split(" = ")[1]
                print "\nERROR Failed to upgrade firmware : " + ERROR_MSG
                sys.exit(EXIT_CODE_CMD_FAILED)
            if "Status = Failure" in textlin:
                FAILURE = True

elif sys.argv[ARG_FUNC_1ST] == GET_INFO:
    if len(sys.argv) <= 2:
        print USAGE_TEXT
        sys.exit(EXIT_CODE_BAD_ARG)

    if sys.argv[ARG_FUNC_2ND] == PHYS:
        # /opt/MegaRAID/storcli/storcli64 /c0/pall show
        cmdline = shlex.split(EXEFILE + " " + CONTROLLER + PHY + " " + GET_INFO)
    elif sys.argv[ARG_FUNC_2ND] == PHY_ERR_COUNTERS:
        # /opt/MegaRAID/storcli/storcli64 /c0/pall show phyerrorcounters
        cmdline = shlex.split(EXEFILE + " " + CONTROLLER + PHY + " " + GET_INFO + " " + STATS)
    else:
        # /opt/MegaRAID/storcli/storcli64 /c0 show
        cmdline = shlex.split(EXEFILE + " " + CONTROLLER + " " + GET_INFO)

    try:
        output = subprocess.check_output(cmdline)
        cont_print = False

        for textlin in output.split("\n"):
            if sys.argv[ARG_FUNC_2ND] == FW_VERSION:
                if "FW Version" in textlin:
                    print(textlin.split(" = ")[1])
            elif sys.argv[ARG_FUNC_2ND] == PRODUCT_NAME:
                if "Product Name" in textlin:
                    print(textlin.split(" = ")[1])
            elif sys.argv[ARG_FUNC_2ND] == SERIAL_NUMBER:
                if "Serial Number" in textlin:
                    print(textlin.split(" = ")[1])
            elif sys.argv[ARG_FUNC_2ND] == PHYS:
                if "---------------------" in textlin:
                    cont_print = True
                    print(textlin)
                elif "Unknown" in textlin:
                    cont_print = True
                elif cont_print:
                    print(textlin)
            elif sys.argv[ARG_FUNC_2ND] == PHY_ERR_COUNTERS:
                if "Phy Info :" in textlin:
                    cont_print = True
                if "PhyNo-Phy Number" in textlin:
                    if len(sys.argv) <= ARG_FUNC_3RD:
                        cont_print= False
                    elif sys.argv[ARG_FUNC_3RD] != VERBOSE:
                        cont_print= False
                if cont_print:
                    print(textlin)
            else:
                print USAGE_TEXT
                sys.exit(EXIT_CODE_BAD_ARG)

    except subprocess.CalledProcessError as e:
        for textlin in e.output.split("\n"):
            # ex. Status = Failure
            #     Description = Controller 0 not found
            if FAILURE:
                ERROR_MSG = textlin.split(" = ")[1]
                print "\nERROR Failed to upgrade firmware : " + ERROR_MSG
                sys.exit(EXIT_CODE_CMD_FAILED)
            if "Status = Failure" in textlin:
                FAILURE = True
else:
    print USAGE_TEXT
    sys.exit(EXIT_CODE_BAD_ARG)
